// We can't parse arbitrary RegExs with regular expressions, because nesting.

// We can lex them, however, and write a small parser on top.

import {atomic, avoid, flags, either, lookBehind, namedCapture, sequence, suffix, lookAhead} from "../compose-regexp.js"

const maybe = suffix("?")

const tag = (name, ...args) => (
    typeof name === 'string' 
    ? namedCapture(name, ...args)
    : either(
        ...Object.entries(name).map(
            ([name, pattern]) => namedCapture(name, pattern)
        )
    )
)

const set = (...items) => new RegExp(`[${items.sort().join("")}]`)

set.complement = set => {
    const {source} = set
    const isNeg = source[1] === "^"
    return isNeg ?
    new RegExp(source.replace(/(?<=^\[)^$/g, ""))
    : new RegExp(source.replace(/(?<=^\[)()/, "^"))
}

const SyntaxCharacter = set(..."^$\\.*+?()[]{}|")
const PatternCharacter = set.complement(SyntaxCharacter)
const ExtendedPatternCharacter = set.complement(set(..."^$\\.*+?()[|"))

const ControlLetter = /[A-Za-z]/
const ControlEscape = set(..."fnrtv")
const HexDigit = set("a-z", "A-Z", "0-9")

const OneNine = /[1-9]/
const DecimalDigit = /[0-9]/

const DecimalEscape = atomic(sequence(OneNine, suffix("*", DecimalDigit)))

const IdStart = ({U, N})=> either(
    /\p{ID_Start}/,
    ['\\']
)

const IdPart = ({U, N})=> either()

const IdName = ({U, N})=> sequence(IdStart({U, N}), suffix("*", IdPart({U, N})))

const AssertionSimple = either(
    "^", "$", "\\b", "\\B"
)

const BraceQuantifier = sequence("{", /\d+/, maybe(","), suffix("*", /\d+/), "}")
const Quantifier = sequence(
    either(
        "+", "*", "?",
        BraceQuantifier
    ),
    maybe("?")
)

const AssertionOpen = either(
    "(?<=",
    "(?<!",
    "(?=",
    "(?!"
)

const GroupOpen = ({U, N}) => either(
    ["(?<", IdName({U, N}), ">"],
    "(?:",
    "("
)

const AssertionAndGroupClose = sequence(")", maybe(tag("GroupOrAssertionQuantifier", Quantifier)))

const CharacterClassOpen = tag("CharacterClassOpen", either("[^", "["))

const CharacterClassClose = tag("CharacterClassClose", "]", maybe(tag("CharacterClassQuantifier", Quantifier)))

function UnicodeEscapeSequence({U, N}) {
    return U
    ? sequence(
        "u",
        either(
            [
                tag("UnicodeEscapeLeadSurrogate", /d[89ABab]/, suffix("{2}", HexDigit)),
                "\\u", 
                tag("UnicodeEscapeTrailSurrogate", /d[CDEFcdef]/, suffix("{2}", HexDigit))
            ],
            maybe("{"),
            tag("UnicodeEscapeSequence", either(
                suffix("{4}", HexDigit),
                [lookBehind("{"), suffix("{1,6}", HexDigit), lookAhead("}")]
            ),
            maybe("}")
        )
    ))
    : sequence('u'), tag("UnicodeEscapeSequence", suffix("{4}", HexDigit))
}

function UnicodeCharacterClassEscape({U,N}) {
    return U
    ? [sequence(tag("UnicodePropertyTag", /[pP]/), "{", tag("UnicodePropertyValue"),"}")]
    : []
}

export const plainOptions = {
    UnicodeCharacterClassEscape: [],
    UnicodeEscapeSequence: []
}

export const unicodeOptions = {
    UnicodeCharacterClassEscape: [],
    UnicodeEscapeSequence: []
}


// ----- // ----- // ----- // ----- // ----- // ----- // ----- // ----- // ----- //



const CharacterClassEscape = ({U, N}) => either(
    tag("PlainCharacterClassEscape", /[dDsSwW]/),
    ...UnicodeCharacterClassEscape({U, N})
)


const HexEscapeSequence = sequence("x", tag("HexEscapeSequence", (suffix("{2}", HexDigit))))

const CharacterEscape = ({U,N}) => either(
    tag({ControlEscape}),
    ['c', tag({ControlLetter})],
    tag("NullEscape", ['0', avoid(DecimalDigit)]),
    HexEscapeSequence,
    UnicodeEscapeSequence({U, N}),
    SyntaxCharacter,
    "/"
)

const AtomEscape = ({U, N}) => sequence('\\', either(
    DecimalEscape,
    CharacterClassEscape({U,N}),
    CharacterEscape({U,N})
))



const SimpleAtom = ({U, N}) => either (
    PatternCharacter,
    ".",
    ["\\", AtomEscape({U, N})]
)

const ExtendedSimpleAtom = ({N}) => {
    const U = false
    return either (
        ".",
        ["\\", AtomEscape({U, N})],
        ["\\", lookAhead('c')],
        tag("InvalidBraceQuantifier", BraceQuantifier),
        ExtendedPatternCharacter
    )
}

const SimpleAssertion = either(/\b/, /\B/)

const SimpleTerm = ({U, N}) => U
? either(
    SimpleAssertion,
    [SimpleAtom({U, N}), maybe(tag("SimpleQuantifier", Quantifier))]
)
: either(
    SimpleAssertion,
    ExtendedSimpleAtom({N})
)

const Termish = ({U, N}) => either(
    tag({AssertionOpen}),
    tag("GroupOpen", GroupOpen({U, N})),
    SimpleTerm({U, N}),
    "|",
    CharacterClassOpen,
    AssertionAndGroupClose
)
const CharClassItem = ({U, N}) => either(
    CharacterClassClose,
)

console.log(Termish({U:true, N: true}))